from datetime import datetime
import json
import platform
import textwrap
import warnings

from .compat import string_types, binary_type
from .exceptions import *
from .timezone import utc
from .version import version


# TODO: use urlfetch for Google App Engine
try:
    import requests
    _session = requests.session()
    _lib = 'requests'
    _lib_ver = requests.__version__
except ImportError:
    warnings.warn(
        '\n\n' + textwrap.fill(
            'requests library is not available, falling to urllib2. '
            'Note that urllib2 does NOT verifies SSL certificates, so '
            'we recommend to install requests, if possible.'
        )
    )
    try:
        from urllib2 import Request, urlopen, HTTPError, URLError, \
            __version__ as urllib_ver
        _lib = 'urllib2'
        _lib_ver = urllib_ver
    except ImportError:
        from urllib.request import Request, urlopen
        from urllib.error import HTTPError, URLError
        _lib = 'urllib2'
        _lib_ver = 'internal'

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


def urlify(params, prefix=''):
    from .models import CubObject

    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False

    result = {}
    items = params.items() if isinstance(params, dict) else enumerate(params)
    for k, v in items:
        key = '%s[%s]' % (prefix, k) if prefix else k
        if isinstance(v, (dict, list, tuple)):
            result.update(urlify(v, key))
        elif isinstance(v, bool):
            result[key] = 'true' if v else 'false'
        elif v is None:
            result[key] = 'null'
        elif isinstance(v, string_types):
            if isinstance(v, binary_type):
                # Must be utf-8, raise exception if it isn't
                v = v.decode('utf-8')
            if v in ('true', 'false', 'null') or is_number(v):
                v = '"%s"' % v
            result[key] = v
        elif isinstance(v, CubObject):
            result[key] = v.id
        else:
            result[key] = v
    return result


def json_datetime_hook(dikt):
    for k, v in dikt.items():
        if isinstance(v, string_types) and v.endswith('Z'):
            try:
                dt = datetime.strptime(v, '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                pass
            else:
                dt = dt.replace(tzinfo=utc)
                dikt[k] = dt
    return dikt


class API(object):
    def __init__(self, key=None, base_url=None, timeout=None):
        from .config import api_key, api_url, api_timeout
        self.api_key = key or api_key
        self.base_url = base_url or api_url
        self.timeout = timeout or api_timeout

    def url(self, url):
        return '%s%s' % (self.base_url, url)

    def requests_request(self, method, url, data, headers, timeout):
        if method == 'get':
            abs_url = '%s?%s' % (url, urlencode(data))
            params = {}
        else:
            abs_url = url
            params = data
        response = _session.request(
            method,
            abs_url,
            data=params,
            headers=headers,
            timeout=timeout
        )
        http_code = response.status_code
        http_body = response.content
        return http_code, http_body

    def urllib2_request(self, method, url, data, headers, timeout):
        params = urlencode(data)
        if method == 'get':
            abs_url = '%s?%s' % (url, params)
            req = Request(abs_url, None, headers)
        elif method == 'post':
            req = Request(url, params.encode('ascii'), headers)
        elif method == 'delete':
            abs_url = '%s?%s' % (url, params)
            req = Request(abs_url, None, headers)
            req.get_method = lambda: 'DELETE'
        else:
            raise APIError('Unsupported method: %s' % method)

        try:
            response = urlopen(req, timeout=timeout)
            http_code = response.code
            http_body = response.read()
        except HTTPError as e:
            http_code = e.code
            http_body = e.read()
        return http_code, http_body

    def request(self, method, url, params=None):
        params = params or {}
        if not self.api_key:
            raise Unauthorized(
                'You did not provide an API key. There are 2 ways to do it:\n'
                '\n1) set it globally for all requests via '
                'cub.config.api_key, like this:\n'
                '\nimport cub.config\n'
                '\ncub.config.api_key = "<your-key>"\n'
                '\n'
                '\n2) pass it to methods which communicate with the API as '
                'keyword argument, like this:\n'
                '\nfrom cub import User\n'
                '\ninv = User.get(api_key="<your-key>", ...)'
            )

        abs_url = self.url(url)

        client_info = {
            'publisher': 'ivelum',
            'platform': platform.platform(),
            'language': 'Python %s' % platform.python_version(),
            'httplib': '%s v%s' % (_lib, _lib_ver)
        }

        headers = {
            'Authorization': 'Bearer %s' % self.api_key,
            'User-Agent': 'Cub Client for Python, v%s' % version,
            'Content-Type': 'application/x-www-form-urlencoded',
            'X-Cub-User-Agent-Info': json.dumps(client_info)
        }

        data = urlify(params)

        # Send request to API and handle communication errors
        if _lib == 'requests':
            try:
                http_code, http_body = self.requests_request(
                    method,
                    abs_url,
                    data=data,
                    headers=headers,
                    timeout=self.timeout
                )
            except requests.RequestException as e:
                raise ConnectionError(
                    'Cannot connect to Cub API using URL %s' % abs_url
                )
        elif _lib == 'urllib2':
            try:
                http_code, http_body = self.urllib2_request(
                    method,
                    abs_url,
                    data=data,
                    headers=headers,
                    timeout=self.timeout
                )
            except URLError as e:
                raise ConnectionError(
                    'Cannot connect to Cub API using URL %s' % abs_url
                )

        try:
            json_body = json.loads(
                http_body.decode('utf-8'),
                object_hook=json_datetime_hook,
            )
        except Exception as e:
            raise APIError(
                'Invalid response from the API: %s' % http_body,
                http_code,
                http_body
            )

        # Handle API errors
        if http_code != 200:
            try:
                error = json_body['error']
                err_desc = error['description']
                err_params = {}
                err_params.update(error.get('params', {}))
            except (KeyError, TypeError, ValueError):
                raise APIError(
                    'Invalid response from the API: %s' % json_body,
                    http_code,
                    http_body,
                    json_body
                )

            if http_code == 400:
                err_msg = err_desc
                if err_params:
                    err_msg += '\nParams:\n' + '\n'.join(
                        ['%s: %s' % (k, v) for k, v in err_params.items()]
                    )
                raise BadRequest(err_msg, http_code, http_body, json_body)
            elif http_code == 401:
                raise Unauthorized(err_desc, http_code, http_body, json_body)
            elif http_code == 403:
                raise Forbidden(err_desc, http_code, http_body, json_body)
            elif http_code == 404:
                raise NotFound(err_desc, http_code, http_body, json_body)
            elif http_code == 405:
                raise MethodNotAllowed(err_desc, http_code, http_body, json_body)
            elif http_code == 500:
                raise InternalError(err_desc, http_code, http_body, json_body)
            elif http_code == 502:
                raise BadGateway(err_desc, http_code, http_body, json_body)
            elif http_code == 503:
                raise ServiceUnavailable(
                    err_desc, http_code, http_body, json_body
                )
            else:
                raise APIError(err_desc, http_code, http_body, json_body)

        return json_body



from .compat import string_types
from .transport import API, json, json_datetime_hook


def objects_from_json(json_content, api_key=None):
    if isinstance(json_content, string_types):
        try:
            content = json.loads(json_content, object_hook=json_datetime_hook)
        except ValueError:
            return json_content
        return objects_from_json(content, api_key)
    elif isinstance(json_content, list):
        return [objects_from_json(obj, api_key) for obj in json_content]
    elif isinstance(json_content, dict):
        obj = json_content.get('object', None)
        if obj:
            id = json_content['id']
            klass = {
                'user': User,
                'organization': Organization,
                'organizationtrust': OrganizationTrust,
                'member': Member,
                'position': Position,
                'memberposition': MemberPosition,
                'group': Group,
                'groupmember': GroupMember,
                'servicesubscription': ServiceSubscription,
                'plan': Plan,
                'customer': Customer,
                'site': Site,
                'mailinglist': MailingList,
                'subscription': Subscription,
                'country': Country,
                'state': State,
                'product': Product,
                'sku': SKU,
                'order': Order,
                'orderitem': OrderItem,
                'charge': Charge,
                'lead': Lead,
                'message': Message,
                'usersite': UserSite,
            }.get(obj, CubObject)
            return klass(api_key=api_key, id=id).load_from(json_content)
        else:
            return json_content


class CubObject(object):

    def __init__(self, api_key=None, id=None, deleted=False, **kwargs):
        self.id = id
        self.api_key = api_key
        self.deleted = deleted
        self._values = {'deleted': deleted}
        self.load_from(kwargs)

    def __setattr__(self, key, value):
        if hasattr(self, '_values'):
            self._values[key] = value
        return super(CubObject, self).__setattr__(key, value)

    def __repr__(self):
        return '%s id=%s: %s' % (self.__class__.__name__, self.id, self._values)

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __eq__(self, other):
        if not hasattr(other, '_values'):
            return False
        if len(self._values) != len(other._values):
            return False
        for k, v in self._values.items():
            if not hasattr(other, k) or self._values[k] != other._values[k]:
                return False
        return True

    @classmethod
    def class_url(cls):
        return '/%ss' % cls.__name__.lower()

    def instance_url(self):
        assert self.id is not None, 'Unable to determine instance URL, '\
                                    'because %s.id is None.' \
                                    '' % self.__class__.__name__
        return '%s/%s' % (self.class_url(), self.id)

    def load_from(self, dikt):
        if dikt is None:
            return self
        for k, v in dikt.items():
            if isinstance(v, (dict, list)):
                v = objects_from_json(v, self.api_key)
            self.__setattr__(k, v)
        return self

    def reload(self, **kwargs):
        response = API(self.api_key).request('get', self.instance_url(), kwargs)
        return self.load_from(response)

    @classmethod
    def get(cls, id, api_key=None, **kwargs):
        instance = cls(api_key=api_key, id=id)
        return instance.reload(**kwargs)


class CreatableObject(CubObject):
    @classmethod
    def create(cls, api_key=None, **kwargs):
        api = API(api_key)
        response = api.request('post', cls.class_url(), params=kwargs)
        return objects_from_json(response, api.api_key)


class UpdatableObject(CubObject):
    def save(self):
        if self.id:
            response = API(self.api_key).request(
                'post',
                self.instance_url(),
                params=self._values
            )
        else:
            response = API(self.api_key).request(
                'post',
                self.__class__.class_url(),
                params=self._values
            )
        return self.load_from(response)


class ListableObject(CubObject):
    @classmethod
    def list(cls, api_key=None, **kwargs):
        response = API(api_key).request('get', cls.class_url(), kwargs)
        return objects_from_json(response, api_key=api_key)


class RemovableObject(CubObject):
    def delete(self):
        response = API(self.api_key).request('delete', self.instance_url())
        return self.load_from(response)

    @classmethod
    def delete_id(cls, id, api_key=None):
        instance = cls(api_key=api_key, id=id)
        return instance.delete()


# -------------- User objects -------------- #


class User(UpdatableObject):
    def load_from(self, dikt):
        if 'token' in dikt:
            self.api_key = dikt['token']
        return super(User, self).load_from(dikt)

    @classmethod
    def class_url(cls):
        return '/user'

    def instance_url(self):
        return '/users/%s' % self.id if self.id else self.class_url()

    @classmethod
    def get(cls, api_key=None, **kwargs):
        instance = cls(api_key=api_key)
        return instance.reload(**kwargs)

    @classmethod
    def login(cls, username, password, provider=None, api_key=None):
        provider = provider or ''
        url = '%s/%s' % (cls.class_url(), 'login')
        response = API(api_key).request('post', url, {
            'username': username,
            'password': password,
            'provider': provider,
        })
        return cls().load_from(response)

    def reissue_token(self, **kwargs):
        url = '%s/%s' % (self.class_url(), 'reissue-token')
        response = API(self.api_key).request('post', url, kwargs)
        return self.load_from(response)


class Organization(ListableObject):
    pass


class OrganizationTrust(CubObject):
    pass


class Position(ListableObject):
    pass


class Member(CreatableObject, UpdatableObject, RemovableObject, ListableObject):
    pass


class MemberPosition(CreatableObject, UpdatableObject,
                     RemovableObject, ListableObject):
    pass


class Group(CreatableObject, UpdatableObject, RemovableObject, ListableObject):
    pass


class GroupMember(CreatableObject, RemovableObject, ListableObject):
    pass


class Plan(CubObject):
    pass


class Customer(CreatableObject, ListableObject):
    pass


class ServiceSubscription(ListableObject):
    pass


class Site(ListableObject):
    pass


class MailingList(ListableObject):
    pass


class Subscription(ListableObject):
    pass


class Country(ListableObject):
    @classmethod
    def class_url(cls):
        return '/countries'


class State(ListableObject):
    pass


class Product(CubObject):
    pass


class SKU(CubObject):
    pass


class Order(CreatableObject, UpdatableObject, RemovableObject, ListableObject):
    pass


class OrderItem(CreatableObject, UpdatableObject, RemovableObject,
                ListableObject):
    pass


class Charge(ListableObject):
    pass


class Lead(ListableObject):
    pass


class Message(ListableObject):
    pass


class UserSite(ListableObject):
    pass



from datetime import timedelta, tzinfo


try:
    import pytz
except ImportError:
    pytz = None


ZERO = timedelta(0)


class UTC(tzinfo):
    """
    UTC implementation taken from Python's docs.

    Used only when pytz isn't available.
    """

    def __repr__(self):
        return '<UTC>'

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return ZERO

utc = pytz.utc if pytz else UTC()



#!/usr/bin/env python

import os
import cub

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

os.environ['PYTHONDONTWRITEBYTECODE'] = '1'

packages = ['cub']
requires = ['requests>=0.9']

setup(
    name='cub',
    version=cub.__version__,
    description='Cub Client for Python',
    long_description=open('README.rst').read(),
    author='Denis Stebunov',
    author_email='support@praetoriandigital.com',
    url='https://github.com/praetoriandigital/cub-python',
    packages=packages,
    python_requires='>=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    install_requires=requires,
    license=open('LICENSE').readline().strip(),
    zip_safe=False,
    test_suite="tests",
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ),
)

del os.environ['PYTHONDONTWRITEBYTECODE']
